###########################################################################

close_events <- function(data, event = c("move", "flipCard", "openTopic", "openPopup")) {

  event <- match.arg(event)

  switch(event,
    "move" = {
      actions <- c("Transform start", "Transform stop")
      idvar   <- c("fileId", "eventId", "artwork", "glossar")
      drop    <- c("popup", "topicNumber", "trace", "event")
      ncol    <- 16

    },
    "flipCard" = {
      actions <- c("Show Info", "Show Front")
      idvar   <- c("fileId", "trace", "artwork", "glossar")
      drop    <- c("popup", "topicNumber", "eventId", "event")
      ncol    <- 16

    },
    "openTopic" = {
      actions <- c("Artwork/OpenCard", "Artwork/CloseCard")
      idvar   <- c("fileId", "eventId", "trace", "glossar", "artwork",
                   "topicNumber")
      drop    <- c("popup", "event")
      ncol    <- 18

    },
    "openPopup" = {
      actions <- c("ShowPopup", "HidePopup")
      idvar   <- c("fileId", "eventId", "trace", "glossar", "artwork", "popup")
      drop    <- c("topicNumber", "event")
      ncol    <- 18
#   TODO: Should topicNumber maybe also be filled in for "openPopup"?

    }
  )

  subdata <- subset(data, data$event %in% actions)
  subdata <- subdata[order(subdata$artwork, subdata$popup, subdata$date, subdata$timeMs), ]
  subdata$time <- ifelse(subdata$event == actions[1], "start", "stop")
  num_start <- diff(c(0, which(subdata$event == actions[2])))
  if (utils::tail(subdata, 1)$time == "start") {
    num_start <- c(num_start, 1)
  }
  subdata$eventId <- rep(seq_along(num_start), num_start)

  if (event == "move") {
    subdata    <- subdata[!duplicated(subdata[, c("event", "eventId")]), ]
    id_stop    <- which(subdata$event == actions[2])
    id_rm_stop <- id_stop[diff(id_stop) == 1]
    subdata    <- subdata[-(id_rm_stop + 1), ]
  }

  subdata_split <- split(subdata, ~ fileId)

  pbapply::pboptions(style = 3, char = "=")

  subdata_split_wide <- pbapply::pblapply(subdata_split, stats::reshape,
                               direction = "wide",
                               idvar = idvar,
                               timevar = "time",
                               drop = drop)
#  suppressWarnings(
#    data_wide <- stats::reshape(subdata, direction = "wide",
#                         idvar = idvar,
#                         timevar = "time",
#                         drop = drop)
#  )

  # remove entries with only start or stop events since they do not have
  # all columns
  ids <- which(sapply(subdata_split_wide, ncol) != ncol)
  if (length(ids) > 0) subdata_split_wide <- subdata_split_wide[-ids]

  data_wide <- dplyr::bind_rows(subdata_split_wide)

  for (d in drop) data_wide[d] <- NA
  data_wide$distance        <- NA
  data_wide$scaleSize       <- NA
  data_wide$rotationDegree  <- NA

  data_wide$event <- event
  data_wide$duration <- data_wide$timeMs.stop - data_wide$timeMs.start

  if (event == "move") {
    data_wide$distance <- apply(
        data_wide[, c("x.start", "y.start", "x.stop", "y.stop")], 1,
        function(x) stats::dist(matrix(x, 2, 2, byrow = TRUE)))
    data_wide$rotationDegree <- data_wide$rotation.stop -
      data_wide$rotation.start
    data_wide$scaleSize <- data_wide$scale.stop / data_wide$scale.start
    # remove moves without any change
    move_wide <- data_wide[data_wide$distance != 0 &
                           data_wide$rotationDegree != 0 &
                           data_wide$scaleSize != 1, ]
    cat(paste("INFORMATION:", nrow(data_wide) - nrow(move_wide),
    "lines containing move events were removed since they did",
    "\nnot contain any change"), fill = TRUE)
    data_wide <- move_wide
  }

  out <- data_wide[, c("fileId", "event", "artwork", "trace", "glossar",
                       "date.start", "date.stop", "timeMs.start",
                       "timeMs.stop", "duration", "topicNumber", "popup",
                       "x.start", "y.start", "x.stop", "y.stop",
                       "distance", "scale.start", "scale.stop",
                       "scaleSize", "rotation.start", "rotation.stop",
                       "rotationDegree")]
  rownames(out) <- NULL
  out
}

