###########################################################################

# Add case variable

add_case <- function(data, cutoff = 20) {
# TODO: What is the best choice for the cutoff here?

  data$timediff <- as.numeric(diff(c(data$date.start[1], data$date.start)))
  data$case <- NA
  j <- 1
  pb <- utils::txtProgressBar(min = 0, max = nrow(data), initial = NA, style = 3)

  for (i in seq_len(nrow(data))) {
    if (data$timediff[i] <= cutoff) {
      data$case[i] <- j
    } else {
      j <- j + 1
      data$case[i] <- j
    }
    utils::setTxtProgressBar(pb, i)
  }
  data$timediff <- NULL
  data
}

###########################################################################

# Add trace for moves

add_trace_moves <- function(data) {

  pbapply::pboptions(style = 3, char = "=")

  trace_max <- max(data$trace, na.rm = TRUE)

  #subdata_art <- split(data, ~ artwork)
  subdata_case <- split(data, ~ case)

  #subdata_list <- split(data, ~ artwork + case)
  # --> does not work with complete data set
  cat("Splitting data...", "\n")
  subdata_list <- pbapply::pblapply(subdata_case, split, f = ~artwork)
  subdata_list <- unlist(subdata_list, recursive = FALSE)

  cat("Adding trace...", "\n")
  subdata_trace <- pbapply::pblapply(subdata_list,
        function(x) {
          trace_max <<- trace_max + 1
          add_trace_subdata(x, max_trace = trace_max)
        }
      )

  out <- dplyr::bind_rows(subdata_trace)
  out <- out[order(out$fileId, out$date.start, out$timeMs.start), ]
  rownames(out) <- NULL

  # Make trace a consecutive number
  out$trace <- as.numeric(factor(out$trace, levels = unique(out$trace)))
  out
}


add_trace_subdata <- function(subdata, max_trace) {

  if (nrow(subdata) != 0) {

    if (length(stats::na.omit(unique(subdata$trace))) == 1) {
      subdata[subdata$event == "move", "trace"] <- stats::na.omit(unique(subdata$trace))
    } else if (length(stats::na.omit(unique(subdata$trace))) > 1) {
      for (i in 1:nrow(subdata)) {
        if (subdata$event[i] == "move") {
          if (i == 1) {
            subdata$trace[i] <- stats::na.omit(unique(subdata$trace))[1]
          } else {
            subdata$trace[i] <- subdata$trace[i - 1]
          }
        }
      }
    } else if (all(is.na(subdata$trace))) {
      for (i in 1:nrow(subdata)) {
        subdata$trace[i] <- max_trace
      }
    }

  } else {
    warning("subdata has nrow = 0")
  }
  subdata
}


###########################################################################

# Create data frame with file names and topics for each artwork

extract_topics <- function(artworks, pattern, path) {

  dat <- NULL
  file_order <- NULL
  i <- 1

  for (artwork in artworks) {

    if (length(pattern) == 1) {
      index_file <- pattern
    } else {
      index_file <- pattern[i]
    }

    fnames <- dir(pattern = paste0(artwork, "_"),
                  path = paste(path, artwork, sep = "/"))
    topic <- NULL
    for (fname in fnames) {
      suppressWarnings(
      topic <- c(topic, gsub("^<card type=.(.*).>$", "\\1",
        grep("^<card type=",
          trimws(readLines(paste(path, artwork, fname, sep = "/"))),
          value = T)))
      )

    }
    index <- paste(path, artwork, index_file, sep = "/")
    suppressWarnings(
    file_order <- c(file_order, gsub("^<card src=.*/(.*)./>$", "\\1",
                         grep("^<card src=", trimws(readLines(index)),
                              value = TRUE)))
    )
    in_index <- fnames %in% file_order
    dat <- rbind(dat, data.frame(artwork, file_name = fnames, in_index, topic))
    i <- i + 1
  }

  # take only the ones that are actually displayed and sort in the same order
  # as indicated in index.html
  out <- dat[dat$in_index, -3]
  out <- out[order(file_order, out$file_name), ]
  rownames(out) <- NULL

  out$index <- unlist(sapply(table(out$artwork), seq_len))
  out
}

###########################################################################

# Add topics: file names and topics

add_topic <- function(data, topics) {

  artworks <- unique(data$artwork)
  tab_art <- lapply(artworks,
               function(x) names(table(data$topicNumber[data$artwork == x])))
  names(tab_art) <- artworks

  tab_index <- lapply(tab_art, seq_along)

  dat_split <- split(data, ~ artwork)

  set_label <- function(x) {
    artwork <- unique(x$artwork)
    x$topicIndex <- factor(x$topicNumber, labels = tab_index[[artwork]])
    x
  }

  dat_label <- lapply(dat_split, set_label)

  set_topic <- function(x) {
    artwork <- unique(x$artwork)
    labels_file <- topics[topics$artwork == artwork,
                          "file_name"][as.numeric(levels(x$topicIndex))]
    x$topicFile <- as.character(factor(x$topicIndex, labels = labels_file))
    labels_topic <- topics[topics$artwork == artwork,
                          "topic"][as.numeric(levels(x$topicIndex))]
    x$topic <- as.character(factor(x$topicIndex, labels = labels_topic))
    x
  }

  dat_topic <- lapply(dat_label, set_topic)

  #out <- do.call(rbind, dat_topic)
  out <- dplyr::bind_rows(dat_topic)
  out$topicIndex <- as.numeric(out$topicIndex)
  out <- out[order(out$fileId, out$date.start, out$timeMs.start), ]
  rownames(out) <- NULL
  out
}

###########################################################################

# Create data frame with information on artworks

extract_artworks <- function(artworks, files = paste0(artworks, ".xml"),
                             path = path) {
  out <- NULL
  i <- 1

  for (artwork in artworks) {

    if (length(files) == 1) {
      index_file <- files
    } else {
      index_file <- files[i]
    }

    index <- paste(path, artwork, index_file, sep = "/")
    varnames <- c("artist", "title", "misc", "description")
    xmllist <- XML::xmlToList(index)$header[varnames]

    if (any(sapply(xmllist, is.null))) {# necessary for missing entries
      names(xmllist) <- varnames
      xmllist[which(sapply(xmllist, is.null))] <- NA
    }
    # remove German quotes
    xmllist <- lapply(xmllist, function(x) gsub("\u201e|\u201c", "", x))
    # remove HTML tags
    xmllist <- lapply(xmllist, function(x) gsub("<br/>", " ", x))
    xmldat <- as.data.frame(xmllist)
    xmldat$artwork <- artwork
    # trim white space from strings
    xmldat$artist      <- trimws(xmldat$artist)
    xmldat$title       <- trimws(xmldat$title)
    xmldat$misc        <- trimws(xmldat$misc)
    xmldat$description <- trimws(xmldat$description)
    out <- rbind(out, xmldat)
    i <- i + 1
  }
  out
}

